-- ===================================================================

-- The Unified Inventory part of this code is:
-- Copyright (c) 2012 CoRNeRNoTe, Dean Montgomery
-- License: GPLv3

-- As this module incorporates GPL3 code,  the entire module is  GPL3.
-- However,  the  rest of the code is  (c) 2017-2019  OldCoder (Robert
-- Kiraly).

-- ===================================================================

-- "miniskins.fallback"  should specify a  64x32 (not 64x64)  fallback
-- skin PNG file that's guaranteed to exist.

local modname = minetest.get_current_modname()
local modpath = minetest.get_modpath (modname)
local storage = minetest.get_mod_storage()

miniskins          = {}
miniskins.pages    = {}
miniskins.fallback = "player_character.png"
miniskins.textdir  = modpath .. "/textures/"

-- ===================================================================

-- This table  maps a player name to an overlay name  if, and only if,
-- a skin has been displayed for the player and the associated overlay
-- was added  (or  an  attempt was made  to add it) in the most recent
-- visual update.

-- In all other cases, the map result is nil.

-- Presently, possible overlay names include:
--
--    spacesuit -- for the spacesuit
--    crystal   -- for crystal armor

miniskins.overlay_name = {}

-- ===================================================================

dofile (modpath .. "/oldmap.lua"   )

-- ===================================================================

local cslog = function (s)
    if false then
        ocutil.log ("[miniskins] " .. s)
    end
end

-- ===================================================================

miniskins.png_exists  = function (basepng)
    return ocutil.mod_texture_exists  (modname, basepng)
end

-- ===================================================================

miniskins.png_missing = function (basepng)
    return ocutil.mod_texture_missing (modname, basepng)
end

-- ===================================================================

local get_stored_trbase = function (plname)
    if  ocutil.str_empty (plname) then
        ocutil.panic ("get_stored_trbase")
    end
    local trbase = storage:get_string (plname)
    if ocutil.str_empty (trbase) then return "" end
    return trbase
end

-- ===================================================================

local get_stored_basepng = function (plname)
    local  trbase = get_stored_trbase (plname)
    if ocutil.str_empty (trbase) then return "" end
    return trbase .. ".png"
end

-- ===================================================================

local set_stored_trbase = function (plname, trbase)
    if  ocutil.str_empty (plname ) or
        ocutil.str_empty (trbase ) then
        ocutil.panic ("set_stored_trbase")
    end

    trbase = trbase:gsub ("%.png$", "")

    if  storage:get_string (plname) ~= trbase then
        storage:set_string (plname, trbase)
    end
end

-- ===================================================================

for plname, trbase in pairs (miniskins.oldmap) do
    local stored = get_stored_trbase (plname)

    if ocutil.str_empty (stored) then
        local basepng = trbase .. ".png"
        if miniskins.png_exists (basepng) then
            set_stored_trbase (plname, trbase)
        end
    end
end

-- ===================================================================

function get_size (w)
    local file = io.open (w)
    if file then
        file:seek("set", 16)
        local widthstr, heightstr = file:read(4), file:read(4)

        local width  = widthstr:sub(1,1):byte()
            * 16777216 + widthstr:sub(2,2):byte()
            * 65536    + widthstr:sub(3,3):byte()
            * 256      + widthstr:sub(4,4):byte()

        local height = heightstr:sub(1,1):byte()
            * 16777216 + heightstr:sub(2,2):byte()
            * 65536    + heightstr:sub(3,3):byte()
            * 256      + heightstr:sub(4,4):byte()

        file:close()
        return width, height
    end
end

-- ===================================================================

local function wstrim (s)
    return s:match "^%s*(.-)%s*$"
end

-- ===================================================================

miniskins.get_player_skin = function (plname)

    local fallback = miniskins.fallback
    local gps      = "miniskins.get_player_skin"
    local basepng  = get_stored_basepng (plname)

    cslog (plname .. ": stored basepng is " .. basepng)

    if  basepng == "" or
        miniskins.png_missing (basepng) then
        basepng =  fallback
    end

    if miniskins.png_missing (basepng) then
        ocutil.panic (gps .. " #1")
    end

    local pngpath = miniskins.textdir .. basepng
    local w, h    = get_size (pngpath)

    if w == nil or w ~= 64 or
       h == nil or (h ~= 32 and h ~= 64) then
       cslog (plname .. ": get_size failed")
       if basepng == fallback then ocutil.panic (gps .. " #2") end
       return fallback, 32
    end

    cslog (plname .. ": found and using file " .. basepng)
    return basepng, h
end

-- ===================================================================

miniskins.get_player_skin32 = function (plname)
    local basepng = miniskins.get_player_skin (plname)
    return "[combine:64x32:0,0=" .. basepng, 32
end

-- ===================================================================

-- If the specified  player  is online,  this routine updates the dis-
-- played skin to match the recorded one.

-- "overlay_name" may be an overlay name or nil. Possible overlay nam-
-- es are listed earlier in this file.

-- If it's an overlay name, this routine attempts to add the specified
-- overlay to the skin displayed.

-- "miniskins.overlay_name"  maps a  player name to an overlay name if
-- and only if  such an attempt was made in the most recent update for
-- the player in question.

-- ===================================================================

miniskins.change_skin = function (plname, overlay_name)

    local player = minetest.get_player_by_name (plname)
    if player == nil then return end

    local show_overlay = false
    if overlay_name ~= nil then
          show_overlay = true
    end

    local basepng = miniskins.get_player_skin (plname)
    local pngpath = miniskins.textdir .. basepng
    local png_width, png_height = get_size (pngpath)

    if png_width == nil then
        cslog ("Warning: invalid PNG file: " .. pngpath)
        return
    end

    if show_overlay then
        if  png_height >  32 then
            basepng    =  miniskins.get_player_skin32 (plname)
            png_height =  32
        end

        basepng = basepng .. "^miniskins_" ..
            overlay_name .. "_overlay.png"
    end
    miniskins.overlay_name [plname] = overlay_name

    if png_width / png_height == 2 then
                                -- rectangle texture; e.g., 64x32
        player:set_properties ({
            textures = { basepng, "ptextures_transparent.png" }
        })
    elseif png_width / png_height == 1 then
                                -- square texture; e.g., 64x64
        player:set_properties ({
            textures = { "ptextures_transparent.png", basepng }
        })
    end
end

-- ===================================================================

local labmap = {}

labmap ["mumbo"       ] = "skin_Mumbo_Jumbo"
labmap ["redminion"   ] = "skin_RedMinionToby"

labmap ["09sharkboy"  ] = "skin_09SharkBoy"
labmap ["apple"       ] = "skin_Apple_Tree"
labmap ["athena"      ] = "skin_Athena"
labmap ["battery"     ] = "skin_Battery"
labmap ["bear"        ] = "skin_Bear"
labmap ["brianna"     ] = "skin_Brianna_Playz"
labmap ["casualgirl"  ] = "skin_Girl_Casual"
labmap ["cat"         ] = "skin_Actual_Cat"
labmap ["chad"        ] = "skin_Chad_Wild_Clay"
labmap ["chara"       ] = "skin_Chara"
labmap ["cola"        ] = "skin_Cola"
labmap ["duck"        ] = "skin_Duck"
labmap ["kata"        ] = "skin_Kata_Red_Hair"
labmap ["panda"       ] = "skin_Cool_Panda"
labmap ["cylon"       ] = "skin_Gold_Cylon"
labmap ["elsa"        ] = "skin_Queen_Elsa"
labmap ["enderfire"   ] = "skin_Fire_Enderman"
labmap ["farmer"      ] = "skin_Farmer_Green"
labmap ["goodwitch"   ] = "skin_Good_Witch"
labmap ["hotdog"      ] = "skin_Hotdog"
labmap ["kermit"      ] = "skin_Business_Kermit"
labmap ["luigi"       ] = "skin_Luigi"
labmap ["mario"       ] = "skin_Mario"
labmap ["moose"       ] = "skin_MooseCraft"
labmap ["owl"         ] = "skin_Owl"
-- labmap ["pie"         ] = "skin_Pie"
labmap ["pizzagirl"   ] = "skin_Pizza_Girl"
labmap ["polarbear"   ] = "skin_Polar_Bear"
labmap ["preston"     ] = "skin_Preston_Playz"
labmap ["purple"      ] = "skin_Purple_Outfit"
labmap ["robocop"     ] = "skin_Robocop"
labmap ["rosalyn"     ] = "skin_Potion_Master"
labmap ["shadowlady"  ] = "skin_LDShadow_Lady"
labmap ["spongebob"   ] = "skin_Spongebob"
labmap ["stampy"      ] = "skin_Stampy_Boots"
labmap ["steampunk"   ] = "skin_Steampunk"
labmap ["superjack"   ] = "skin_Super_Jack"
labmap ["superpollo"  ] = "skin_Superpollo"
labmap ["squirrel"    ] = "skin_Squirrel"
labmap ["tiger"       ] = "skin_Tiger"
labmap ["tree"        ] = "skin_Apple_Tree"
labmap ["unspeakable" ] = "skin_Unspeakable"
labmap ["wolfboy"     ] = "skin_Red_Wolf_Boy"
labmap ["apple"       ] = "skin_Apple"
labmap ["banana"      ] = "skin_Banana"
labmap ["orange"      ] = "skin_Orange"
labmap ["potato"      ] = "skin_Potato"

-- ===================================================================

for skin_label, trbase in pairs (labmap) do
    local basepng = trbase .. ".png"
    if not miniskins.png_exists (basepng) then
        cslog ("Dropping skin label " .. skin_label)
        labmap [skin_label] = nil
    end
end

local label_to_trbase = labmap
local trbase_to_label = ocutil.swap_key_value (labmap)
local sorted_labels   = ocutil.ktable_to_vtable (labmap)
labmap = nil

-- ===================================================================

for key, value in pairs (sorted_labels   ) do
    cslog ("sorted_labels: "    .. key .. " => " .. value)
end

for key, value in pairs (trbase_to_label ) do
    cslog ("trbase_to_label: " .. key .. " => " .. value)
end

-- ===================================================================

local prompt = "Use e.g. /skin superjack" ..
    " to choose one of these:"

-- ===================================================================

local param_skins = {
    description = "list or change your own skins" ,
    params      = "skin name (optional)"          ,
    privs       = {}                              ,

    func = function (plname, params)

        local print_prompt = true
        local skin_label  = params
        local trbase

        if ocutil.str_empty (skin_label) then skin_label = "" end
        skin_label = skin_label:gsub ("%.png$", "")

        if ocutil.str_empty (skin_label) then
            trbase = get_stored_trbase (plname)

            if  ocutil.str_nonempty (trbase) then
                cslog ("player " .. plname .. " has trbase " ..
                    trbase)

                local old_label = trbase_to_label [trbase]
                if ocutil.str_empty (old_label) then
                      old_label = trbase
                end

                if ocutil.str_nonempty (old_label) then
                    print_prompt = false
                    minetest.chat_send_player (plname,
"Presently, your skin is " .. old_label .. ". " .. prompt)
                end
            end

            if print_prompt then
                minetest.chat_send_player (plname, prompt)
            end

            local buf    = ""
            local npline = 13
            local noline =  0

            for _, skin_label in pairs (sorted_labels) do
                buf    = buf .. " " .. skin_label
                noline = noline + 1

                if noline >= npline then
                    minetest.chat_send_player (plname, buf)
                    buf    = ""
                    noline = 0
                end
            end

            if noline > 0 then
                minetest.chat_send_player (plname, buf)
            end
            return
        end

        trbase = label_to_trbase [skin_label]

        if  trbase == nil then
            local xp = skin_label:gsub ("%.png$", "") .. ".png"
            if miniskins.png_missing (xp) then
                xp = "skin_" .. xp
                if miniskins.png_missing (xp) then
                    xp = xp:gsub ("^skin_", "player_")
                    if miniskins.png_missing (xp) then
                        minetest.chat_send_player (plname,
"Not a valid skin label")
                        return
                    end
                end
            end
            trbase = xp:gsub ("%.png$", "")
        end

        if miniskins.png_missing (trbase .. ".png") then
            minetest.chat_send_player (plname,
"Sorry, PNG file is missing")
            return
        end

        local oname = miniskins.overlay_name [plname]
        set_stored_trbase (plname, trbase)
        miniskins.change_skin (plname, oname)

        minetest.chat_send_player (plname,
"O.K. your skin has been changed to " .. skin_label)
    end
}

minetest.register_chatcommand ("skin"  , param_skins )
minetest.register_chatcommand ("skins" , param_skins )

-- ===================================================================

minetest.register_privilege ("skin", {
    description = "Can change skins for other players" ,
    give_to_singleplayer=true
})

-- ===================================================================

local param_playerskins = {
    description = "list or change skins for a player" ,
    params      = "<player> <skin_label>"             ,
    privs       = { skin=true }                       ,

    func = function (adname, params)

        local plname, label = string.match (params,
            "^([^ ]+) *(.*)")

        if not plname then plname = params end
        if ocutil.str_empty (plname) then plname = "" end
        if ocutil.str_empty (label ) then label  = "" end

        plname = plname:gsub (" +$", "")        
        label  = label:gsub  (" +$", "")        

        if ocutil.str_empty (plname) then
            core.chat_send_player (adname,
                "Usage: /pskin player_name skin_label")
            return true
        end

        if not core.get_auth_handler().get_auth (plname) then
            core.chat_send_player (adname,
                "Error: No such player: " .. plname)
            return true
        end

        local print_prompt = true
        local skin_label   = label
        local trbase

        if ocutil.str_empty (skin_label) then skin_label = "" end
        skin_label = skin_label:gsub ("%.png%", "")

        if ocutil.str_empty (skin_label) then
            trbase = get_stored_trbase (plname)

            if  ocutil.str_nonempty (trbase) then
                cslog ("player " .. plname .. " has trbase " ..
                    trbase)

                local old_label = trbase_to_label [trbase]
                if ocutil.str_empty (old_label) then
                      old_label = trbase
                end

                if ocutil.str_nonempty (old_label) then
                    print_prompt = false
                    minetest.chat_send_player (adname,
"Presently, their skin is " .. old_label .. ". " .. prompt)
                end
            end

            if print_prompt then
                minetest.chat_send_player (adname, prompt)
            end

            local buf    = ""
            local npline = 13
            local noline =  0

            for _, skin_label in pairs (sorted_labels) do
                buf    = buf .. " " .. skin_label
                noline = noline + 1

                if noline >= npline then
                    minetest.chat_send_player (adname, buf)
                    buf    = ""
                    noline = 0
                end
            end

            if noline > 0 then
                minetest.chat_send_player (adname, buf)
            end
            return
        end

        trbase = label_to_trbase [skin_label]

        if  trbase == nil then
            local xp = skin_label:gsub ("%.png$", "") .. ".png"
            if miniskins.png_missing (xp) then
                xp = "skin_" .. xp
                if miniskins.png_missing (xp) then
                    xp = xp:gsub ("^skin_", "player_")
                    if miniskins.png_missing (xp) then
                        minetest.chat_send_player (adname,
"Not a valid skin label")
                        return
                    end
                end
            end
            trbase = xp:gsub ("%.png$", "")
        end

        if miniskins.png_missing (trbase .. ".png") then
            minetest.chat_send_player (adname,
"Sorry, PNG file is missing")
            return
        end

        local oname = miniskins.overlay_name [plname]
        set_stored_trbase (plname, trbase)
        miniskins.change_skin (plname, oname)

        minetest.chat_send_player (adname,
"O.K. their skin has been changed to " .. skin_label)
    end
}

minetest.register_chatcommand ("playerskin" , param_playerskins )
minetest.register_chatcommand ("pskin"      , param_playerskins )

-- ===================================================================

minetest.register_on_joinplayer (function (player)

    local plname  = player:get_player_name()
    local basepng = "player_" .. plname .. ".png"
    local final   = nil

-- ===================================================================

    miniskins.overlay_name [plname] = nil

-- ===================================================================

    if miniskins.png_exists (basepng) then
        cslog (plname .. ": " .. basepng .. " exists")
        final = basepng
    else
        cslog (plname .. ": " .. basepng .. " doesn't exist")
    end

-- ===================================================================

    if final == nil then
        basepng = get_stored_basepng (plname)

        if ocutil.str_empty (basepng) then
            cslog (plname .. " doesn't have a recorded skin")
        else
            cslog (plname .. " has a recorded skin: " .. basepng)

            if miniskins.png_exists (basepng) then
                cslog (plname .. ": " .. basepng .. " exists")
                final = basepng
            else
                cslog (plname .. ": " .. basepng .. " doesn't exist")
            end
        end
    end

-- ===================================================================

    if final == nil then
        local rslist = {
"Actual_Cat"        ,
"Apple_Tree"        ,
"Battery"           ,
"Bear"              ,
"Cola"              ,
"Fox"               ,
"Heather_Elf"       ,
"Lego_Person"       ,
"Lego_Spaceman"     ,
"Marley_Xia"        ,
"Puppy"             ,
"Rabbit_Coco"       ,
"Squirrel"          ,
"Steampunk"         ,
"Superpollo"        ,
"Tiger"             ,
        }

        cslog ("Setting initial skin for " .. plname)

        local ii       = math.random (#rslist)
        local charname = rslist [ii]
        basepng        = "skin_" .. charname .. ".png"
        cslog (plname .. ": Selected " .. basepng .. " at random")

        if miniskins.png_exists (basepng) then
            cslog (plname .. ": " .. basepng .. " exists")
            final = basepng
        else
            cslog (plname .. ": " .. basepng .. " doesn't exist")
        end
    end

-- ===================================================================

    if final == nil then
        basepng = miniskins.fallback
        cslog (plname .. ": Falling back to " .. basepng)

        if miniskins.png_exists (basepng) then
            cslog (basepng .. " exists")
            final = basepng
        else
            ocutil.panic (basepng .. " doesn't exist")
        end
    end

-- ===================================================================

    cslog ("Final skin for " .. plname .. " is " .. final)
    set_stored_trbase (plname, final)
    miniskins.change_skin (plname)
end)

-- ===================================================================
-- End of file.
